;;; sdiff --- S-expression diff
;;; Copyright © 2021 Arun Isaac <arunisaac@systemreboot.net>
;;;
;;; This file is part of sdiff.
;;;
;;; sdiff is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; sdiff is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with sdiff.  If not, see <https://www.gnu.org/licenses/>.

(define-module (matrix)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)
  #:use-module (ice-9 match)
  #:export (matrix-fold
            matrix-any
            matrix-min
            matrix-row
            matrix-col
            for-matrix-row
            for-matrix-col
            matrix-print))

(define (matrix-fold proc init array)
  (match (array-shape array)
    (((min-row max-row) (min-col max-col))
     (fold (lambda (row result)
             (fold (lambda (col result)
                     (proc row col (array-ref array row col) result))
                   result
                   (iota (1+ (- max-col min-col)) min-col)))
           init
           (iota (1+ (- max-row min-row)) min-row)))))

;; TODO: This is an inefficient implementation. Exit early once result
;; is set to a truthy value.
(define (matrix-any pred matrix)
  (matrix-fold (lambda (i j element result)
                 (or result
                     (pred i j element)))
               #f
               matrix))

(define* (matrix-min matrix #:optional (min-proc min))
  (match (array-shape matrix)
    (((min-row max-row) (min-col max-col))
     (matrix-fold (lambda (i j element result)
                    (min-proc element result))
                  (array-ref matrix min-row min-col)
                  matrix))))

(define (matrix-row matrix row-index)
  (match (array-dimensions matrix)
    ((_ cols)
     (make-shared-array matrix list (list row-index row-index) cols))))

(define (matrix-col matrix col-index)
  (match (array-dimensions matrix)
    ((rows _)
     (make-shared-array matrix list rows (list col-index col-index)))))

(define (for-matrix-row proc matrix)
  (match (array-dimensions matrix)
    ((rows _)
     (for-each (lambda (row-index)
                 (proc row-index (matrix-row matrix row-index)))
               (iota rows)))))

(define (for-matrix-col proc matrix)
  (match (array-dimensions matrix)
    ((_ cols)
     (for-each (lambda (col-index)
                 (proc col-index (matrix-col matrix col-index)))
               (iota cols)))))

(define (matrix-print matrix)
  (for-matrix-row (lambda (_ row)
                    (array-for-each (lambda (element)
                                      (format #t "~a~/" element))
                                    row)
                    (newline))
                  matrix))
